@@ -5,7 +5,10 @@
 # distribution of this software and related documentation without an express
 # license agreement from NVIDIA CORPORATION is strictly prohibited.
 
+from __future__ import annotations
+
 import os
+import re
 import sys
 import imp
 import ast
@@ -51,6 +54,86 @@ builtin_operators[ast.LtE] = "<="
 builtin_operators[ast.Eq] = "=="
 builtin_operators[ast.NotEq] = "!="
 
+class StructInstance:
+    def __init__(self, struct: Struct):
+        self.__dict__['_struct_'] = struct
+        self.__dict__['_c_struct_'] = struct.ctype()
+
+    def __setattr__(self, name, value):
+        assert name in self._struct_.vars, "invalid struct member variable {}".format(name)
+        if isinstance(self._struct_.vars[name].type, array):
+            # wp.array
+            assert isinstance(value, array)
+            assert value.dtype == self._struct_.vars[name].type.dtype, "assign to struct member variable {} failed, expected type {}, got type {}".format(name, self._struct_.vars[name].type.dtype, value.dtype)
+            setattr(self._c_struct_, name, value.__ctype__())
+        elif issubclass(self._struct_.vars[name].type, ctypes.Array):
+            # array type e.g. vec3
+            setattr(self._c_struct_, name, self._struct_.vars[name].type(*value))
+        else:
+            # primitive type
+            setattr(self._c_struct_, name, self._struct_.vars[name].type._type_(value))
+
+        self.__dict__[name] = value
+
+    def __repr__(self):
+        lines = []
+        for k, v in self.__dict__.items():
+            if not k.startswith('_'):
+                lines.append(' ' * 4 + f"{k}: {v.__repr__()}")
+        return "StructInstance(\n" + "\n".join(lines) + "\n)\n"
+
+
+class Struct:
+    def __init__(self, cls, key, module):
+        self.cls = cls
+        self.module = module
+        self.key = key
+
+        self.vars = {}
+        for label, type in self.cls.__annotations__.items():
+            if type == float:
+                type = float32
+            elif type == int:
+                type = int32
+            self.vars[label] = Var(label, type)
+
+        fields = []
+        for label, var in self.vars.items():
+            if isinstance(var.type, array):
+                fields.append((label, array_t))
+            elif issubclass(var.type, ctypes.Array):
+                fields.append((label, var.type))
+            else:
+                fields.append((label, var.type._type_))
+
+        class StructType(ctypes.Structure):
+            _fields_ = fields
+
+        self.ctype = StructType
+
+        if (module):
+            module.register_struct(self)
+
+    def __call__(self):
+        '''
+        This function returns s = StructInstance(self)
+        s uses self.cls as template.
+        To enable autocomplete on s, we inherit from self.cls.
+        For example,
+
+        @wp.struct
+        class A:
+            # annotations
+            ...
+
+        The type annotations are inherited in A(), allowing autocomplete in kernels
+        '''
+        # return StructInstance(self)
+
+        class NewStructInstance(self.cls, StructInstance):
+            def __init__(inst):
+                StructInstance.__init__(inst, self)
+        return NewStructInstance()
 
 class Var:
     def __init__(self, label, type, requires_grad=False, constant=None):
@@ -73,6 +156,8 @@ class Var:
         if (isinstance(self.type, array)):
             #return str(self.type.dtype.__name__) + "*"
             return f"array_t<{str(self.type.dtype.__name__)}>"
+        if (isinstance(self.type, Struct)):
+            return self.type.cls.__name__
         else:
             return str(self.type.__name__)
 
@@ -102,34 +187,51 @@ class Adjoint:
     def __init__(adj, func):
 
         adj.func = func
+        
+        # build AST from function object
+        adj.source = inspect.getsource(func)
+
+        # get source code lines and line number where function starts
+        adj.raw_source, adj.fun_lineno = inspect.getsourcelines(func)
+
+        # keep track of line number in function code
+        adj.lineno = None
+        
+        # ensures that indented class methods can be parsed as kernels
+        adj.source = textwrap.dedent(adj.source)
+        
+        # extract name of source file
+        adj.filename = inspect.getsourcefile(func) or "unknown source file"
+
+        # build AST
+        adj.tree = ast.parse(adj.source)                
+
+        # parse argument types
+        adj.arg_types = typing.get_type_hints(func)
+        adj.args = []
+
+        for name, type in adj.arg_types.items():
+
+            # skip return hint
+            if name == "return":
+                continue
+            
+            # add variable for argument
+            arg = Var(name, type, False)
+            adj.args.append(arg)
+
+
+    # generate function ssa form and adjoint
+    def build(adj, builder):
+
+        adj.builder = builder
 
         adj.symbols = {}            # map from symbols to adjoint variables
         adj.variables = []          # list of local variables (in order)
-        adj.args = []               # list of function arguments (in order)
 
         adj.cond = None             # condition variable if in branch
         adj.return_var = None       # return type for function or kernel
 
-        # build AST from function object
-        adj.source = inspect.getsource(func)
-        
-        # ensures that indented class methods can be parsed as kernels
-        adj.source = textwrap.dedent(adj.source)    
-
-        # build AST
-        adj.tree = ast.parse(adj.source)
-
-        # parse argument types
-        arg_types = typing.get_type_hints(func)
-
-        # add variables and symbol map for each argument
-        for name, t in arg_types.items():
-            adj.symbols[name] = Var(name, t, False)
-
-        # build ordered list of args
-        for a in adj.tree.body[0].args.args:
-            adj.args.append(adj.symbols[a.arg])
-
         # blocks
         adj.blocks = [Block()]
 
@@ -139,12 +241,9 @@ class Adjoint:
         # used to generate new label indices
         adj.label_count = 0
 
-    # generate function ssa form and adjoint
-    def build(adj, builtin_fuctions, user_functions, options):
-
-        adj.builtin_functions = builtin_fuctions
-        adj.user_functions = user_functions
-        adj.options = options
+        # update symbol map for each argument
+        for a in adj.args:
+            adj.symbols[a.label] = a
 
         # recursively evaluate function body
         adj.eval(adj.tree.body[0])
@@ -268,15 +367,11 @@ class Adjoint:
 
     def add_call(adj, func, inputs, min_outputs=None):
 
-        # promote func to list
-        if isinstance(func, list) == False:
-            func = [func,]
-
         # if func is overloaded then perform overload resolution here
         # we validate argument types before they go to generated native code
         resolved_func = None
 
-        for f in func:
+        for f in func.overloads:
             match = True
 
             # skip type checking for variadic functions
@@ -335,12 +430,17 @@ class Adjoint:
                 if isinstance(x, warp.context.Function):
                     arg_types += "function" + ", "
 
-            raise Exception(f"Couldn't find function overload for '{func[0].key}' that matched inputs with types: {arg_types}")
+            raise Exception(f"Couldn't find function overload for '{func.key}' that matched inputs with types: {arg_types}")
 
         else:
             func = resolved_func
 
 
+        # if it is a user-function then build it recursively
+        if not func.is_builtin():
+            adj.builder.build_function(func)
+
+        # evaluate the function type based on inputs
         value_type = func.value_func(inputs)
 
         # handle expression (zero output), e.g.: void do_something();
@@ -359,7 +459,7 @@ class Adjoint:
 
             return None
 
-        # handle multiple value function
+        # handle multiple value functions
         elif (isinstance(value_type, list)):
 
             output = [adj.add_var(v) for v in value_type]
@@ -449,7 +549,7 @@ class Adjoint:
         adj.add_forward(f"if (iter_cmp(var_{iter}) == 0) goto for_end_{cond_block.label};")
 
         # evaluate iter
-        val = adj.add_call(adj.builtin_functions["iter_next"], [iter])
+        val = adj.add_call(warp.context.builtin_functions["iter_next"], [iter])
 
         adj.begin_block()
 
@@ -569,6 +669,8 @@ class Adjoint:
     def eval(adj, node):
 
         try:
+            if hasattr(node, "lineno"):
+                adj.set_lineno(node.lineno-1)                
 
             # top level entry point for a function
             if (isinstance(node, ast.FunctionDef)):
@@ -614,7 +716,7 @@ class Adjoint:
 
                     if var1 != var2:
                         # insert a phi function that selects var1, var2 based on cond
-                        out = adj.add_call(adj.builtin_functions["select"], [cond, var1, var2])
+                        out = adj.add_call(warp.context.builtin_functions["select"], [cond, var1, var2])
                         adj.symbols[sym] = out
 
 
@@ -641,7 +743,7 @@ class Adjoint:
                     if var1 != var2:
                         # insert a phi function that selects var1, var2 based on cond
                         # note the reversed order of vars since we want to use !cond as our select
-                        out = adj.add_call(adj.builtin_functions["select"], [cond, var2, var1])
+                        out = adj.add_call(warp.context.builtin_functions["select"], [cond, var2, var1])
                         adj.symbols[sym] = out
 
 
@@ -719,12 +821,23 @@ class Adjoint:
 
                 if key in adj.symbols:
                     return adj.symbols[key]
-                else:
-                    obj = attribute_to_val(node, adj.func.__globals__)
-                    if not isinstance(obj, warp.constant):
-                        raise TypeError(f"'{key}' is not a local variable or of type warp.constant")
-                    out = adj.add_constant(obj.val)
+                elif isinstance(node.value, ast.Name) and node.value.id in adj.symbols:
+                    # access struct attribute
+                    out = Var(key, adj.symbols[node.value.id].type.vars[node.attr].type)
                     adj.symbols[key] = out
+                    return adj.symbols[key]
+                else:
+
+                    # try and resolve to either a wp.constant
+                    # or a wp.func object
+                    obj = attribute_to_val(node, adj.func.__globals__)
+                    
+                    if isinstance(obj, warp.constant):
+                        out = adj.add_constant(obj.val)
+                        adj.symbols[key] = out          # if referencing a constant
+                    else:
+                        raise TypeError(f"'{key}' is not a local variable, warp function, or warp constant")
+
                     return out
 
             elif (isinstance(node, ast.Str)):
@@ -751,7 +864,7 @@ class Adjoint:
                 right = adj.eval(node.right)
 
                 name = builtin_operators[type(node.op)]
-                func = adj.builtin_functions[name]
+                func = warp.context.builtin_functions[name]
 
                 out = adj.add_call(func, [left, right])
                 return out
@@ -761,7 +874,7 @@ class Adjoint:
                 arg = adj.eval(node.operand)
 
                 name = builtin_operators[type(node.op)]
-                func = adj.builtin_functions[name]
+                func = warp.context.builtin_functions[name]
 
                 out = adj.add_call(func, [arg])
                 return out
@@ -792,7 +905,7 @@ class Adjoint:
                             raise Exception("Error mutating a constant {} inside a dynamic loop, use the following syntax: pi = float(3.141) to declare a dynamic variable".format(sym))
                         
                         # overwrite the old variable value (violates SSA)
-                        adj.add_call(adj.builtin_functions["copy"], [var1, var2])
+                        adj.add_call(warp.context.builtin_functions["copy"], [var1, var2])
 
                         # reset the symbol to point to the original variable
                         adj.symbols[sym] = var1
@@ -853,7 +966,7 @@ class Adjoint:
 
                         # test if we're above max unroll count
                         max_iters = abs(end-start)//abs(step)
-                        max_unroll = adj.options["max_unroll"]
+                        max_unroll = adj.builder.options["max_unroll"]
 
                         if max_iters > max_unroll:
 
@@ -905,7 +1018,7 @@ class Adjoint:
                                 raise Exception("Error mutating a constant {} inside a dynamic loop, use the following syntax: pi = float(3.141) to declare a dynamic variable".format(sym))
                             
                             # overwrite the old variable value (violates SSA)
-                            adj.add_call(adj.builtin_functions["copy"], [var1, var2])
+                            adj.add_call(warp.context.builtin_functions["copy"], [var1, var2])
 
                             # reset the symbol to point to the original variable
                             adj.symbols[sym] = var1
@@ -918,23 +1031,27 @@ class Adjoint:
             elif (isinstance(node, ast.Call)):
 
                 name = None
+                
+                # resolve path (e.g.: module.submodule.attr) expression to a list of module names
+                path = adj.resolve_path(node.func)
+                
+                try:
+                    # try and look up path in function globals
+                    func = eval(".".join(path), adj.func.__globals__)
 
-                # determine if call is to a builtin (e.g.: wp.cos(x)), or to a free-func, e.g.: my_func(x)
-                if (isinstance(node.func, ast.Attribute)):
-                    name = node.func.attr
-                elif (isinstance(node.func, ast.Name)):
-                    name = node.func.id
+                    if isinstance(func, warp.context.Function) == False:
+                        raise RuntimeError()
+                        
+                except Exception as e:
 
-                # built in function
-                if name in adj.builtin_functions:
-                    func = adj.builtin_functions[name]
-
-                # user-defined function in this module
-                elif name in adj.user_functions:
-                    func = adj.user_functions[name]
-
-                else:
-                    raise KeyError("Could not find function {}".format(name))
+                    # try and lookup in builtins, this allows users to avoid 
+                    # using "wp." prefix, and also handles type constructors
+                    # e.g.: wp.vec3 which aren't explicitly a function object
+                    attr = path[-1]
+                    if attr in warp.context.builtin_functions:
+                        func = warp.context.builtin_functions[attr]
+                    else:
+                        raise RuntimeError(f"Could not find function {'.'.join(path)} as a built-in or user-defined function. Note that user functions must be annotated with a @wp.func decorator to be called from a kernel.")
 
                 args = []
 
@@ -965,13 +1082,13 @@ class Adjoint:
                 indices = []
 
                 if isinstance(node.slice, ast.Tuple):
-                    # handles the x[i, j] case (Python 3.8.x upward)
+                    # handles the x[i,j] case (Python 3.8.x upward)
                     for arg in node.slice.elts:
                         var = adj.eval(arg)
                         indices.append(var)
 
                 elif isinstance(node.slice, ast.Index) and isinstance(node.slice.value, ast.Tuple):
-                    # handles the x[i, j] case (Python 3.7.x)
+                    # handles the x[i,j] case (Python 3.7.x)
                     for arg in node.slice.value.elts:
                         var = adj.eval(arg)
                         indices.append(var)
@@ -982,17 +1099,16 @@ class Adjoint:
 
                 if isinstance(target.type, array):
                     
-                    # handles array loads
                     if len(indices) == target.type.ndim:
-                        # handles array loads
-                        out = adj.add_call(adj.builtin_functions["load"], [target, *indices])
+                        # handles array loads (where each dimension has an index specified)
+                        out = adj.add_call(warp.context.builtin_functions["load"], [target, *indices])
                     else:
-                        # handles array views
-                        out = adj.add_call(adj.builtin_functions["view"], [target, *indices])
+                        # handles array views (fewer indices than dimensions)
+                        out = adj.add_call(warp.context.builtin_functions["view"], [target, *indices])
 
                 else:
                     # handles non-array type indexing, e.g: vec3, mat33, etc
-                    out = adj.add_call(adj.builtin_functions["index"], [target, *indices])
+                    out = adj.add_call(warp.context.builtin_functions["index"], [target, *indices])
 
                 return out
 
@@ -1022,7 +1138,7 @@ class Adjoint:
                     
                     for name, rhs in zip(names, out):
                         if (name in adj.symbols):
-                            if (rhs.type != adj.symbols[name].type):
+                            if not types_equal(rhs.type, adj.symbols[name].type):
                                 raise TypeError("Error, assigning to existing symbol {} ({}) with different type ({})".format(name, adj.symbols[name].type, rhs.type))
 
                         adj.symbols[name] = rhs
@@ -1055,7 +1171,7 @@ class Adjoint:
                         indices.append(var)                    
 
                     if (isinstance(target.type, array)):
-                        adj.add_call(adj.builtin_functions["store"], [target, *indices, value])
+                        adj.add_call(warp.context.builtin_functions["store"], [target, *indices, value])
                     else:
                         raise RuntimeError("Can only subscript assign array types")
 
@@ -1078,7 +1194,7 @@ class Adjoint:
                     # handle simple assignment case (a = b), where we generate a value copy rather than reference
                     if isinstance(node.value, ast.Name):
                         out = adj.add_var(rhs.type)
-                        adj.add_call(adj.builtin_functions["copy"], [out, rhs])
+                        adj.add_call(warp.context.builtin_functions["copy"], [out, rhs])
                     else:
                         out = rhs
 
@@ -1110,7 +1226,7 @@ class Adjoint:
 
                 # lookup 
                 name = builtin_operators[type(node.op)]
-                func = adj.builtin_functions[name]
+                func = warp.context.builtin_functions[name]
 
                 out = adj.add_call(func, [left, right])
 
@@ -1140,6 +1256,32 @@ class Adjoint:
             raise
 
 
+    # helper to evaluate expressions of the form
+    # obj1.obj2.obj3.attr in the function's global scope
+    def resolve_path(adj, node):
+
+        modules = []
+
+        while isinstance(node, ast.Attribute):
+            modules.append(node.attr)
+            node = node.value
+
+        if (isinstance(node, ast.Name)):
+            modules.append(node.id)
+
+        # reverse list since ast presents it backward order
+        return [*reversed(modules)]
+
+    # annotate generated code with the original source code line
+    def set_lineno(adj, lineno):
+        if adj.lineno is None or adj.lineno != lineno:
+            line = lineno + adj.fun_lineno
+            source = adj.raw_source[lineno].strip().ljust(70)
+            adj.add_forward(f'// {source}       <L {line}>')
+            adj.add_reverse(f'// adj: {source}  <L {line}>')
+        adj.lineno = lineno
+        
+
 #----------------
 # code generation
 
@@ -1153,7 +1295,6 @@ cpu_module_header = '''
 #define int(x) cast_int(x)
 #define adj_int(x, adj_x, adj_ret) adj_cast_int(x, adj_x, adj_ret)
 
-
 using namespace wp;
 
 '''
@@ -1173,28 +1314,40 @@ using namespace wp;
 
 '''
 
+struct_template = '''
+struct {name}
+{{
+{struct_body}
+}};
+
+'''
+
 cpu_function_template = '''
+// {filename}:{lineno}
 static {return_type} {name}({forward_args})
 {{
-    {forward_body}
+{forward_body}
 }}
 
+// {filename}:{lineno}
 static void adj_{name}({reverse_args})
 {{
-    {reverse_body}
+{reverse_body}
 }}
 
 '''
 
 cuda_function_template = '''
+// {filename}:{lineno}
 static CUDA_CALLABLE {return_type} {name}({forward_args})
 {{
-    {forward_body}
+{forward_body}
 }}
 
+// {filename}:{lineno}
 static CUDA_CALLABLE void adj_{name}({reverse_args})
 {{
-    {reverse_body}
+{reverse_body}
 }}
 
 '''
@@ -1209,7 +1362,7 @@ extern "C" __global__ void {name}_cuda_kernel_forward({forward_args})
 
     set_launch_bounds(dim);
 
-    {forward_body}
+{forward_body}
 }}
 
 
@@ -1221,7 +1374,7 @@ extern "C" __global__ void {name}_cuda_kernel_backward({reverse_args})
 
     set_launch_bounds(dim);
 
-    {reverse_body}
+{reverse_body}
 }}
 
 '''
@@ -1230,12 +1383,12 @@ cpu_kernel_template = '''
 
 void {name}_cpu_kernel_forward({forward_args})
 {{
-    {forward_body}
+{forward_body}
 }}
 
 void {name}_cpu_kernel_backward({reverse_args})
 {{
-    {reverse_body}
+{reverse_body}
 }}
 
 '''
@@ -1344,6 +1497,23 @@ def indent(args, stops=1):
     #return sep + args.replace(", ", "," + sep)
     return sep.join(args)
 
+# generates a C function name based on the python function name
+def make_func_name(func):
+    return re.sub('[^0-9a-zA-Z_]+', '', func.__qualname__.replace('.', '__'))
+
+def codegen_struct(struct, indent=4):
+    body = []
+    indent_block = " " * indent
+    for label, var in struct.vars.items():
+        assert not (isinstance(var.type, Struct))
+
+        body.append(var.ctype() + " " + label + ";\n")
+
+    return struct_template.format(
+        name=struct.cls.__name__,
+        struct_body="".join([indent_block + l for l in body])
+    )
+
 
 def codegen_func_forward_body(adj, device='cpu', indent=4):
     body = []
@@ -1473,12 +1643,14 @@ def codegen_func(adj, device='cpu'):
     else:
         raise ValueError("Device {} is not supported".format(device))
 
-    s = template.format(name=adj.func.__name__,
+    s = template.format(name=make_func_name(adj.func),
                         return_type=return_type,
                         forward_args=indent(forward_args),
                         reverse_args=indent(reverse_args),
                         forward_body=forward_body,
-                        reverse_body=reverse_body)
+                        reverse_body=reverse_body,
+                        filename=adj.filename,
+                        lineno=adj.fun_lineno)
 
     return s
 
