@@ -7,10 +7,9 @@
 
 import os
 
-version = "0.2.0"
+version = "0.3.1"
 
 cuda_path = None        # path to local CUDA toolchain, if None at init time warp will attempt to find the SDK using CUDA_PATH env var
-host_path = None        # path to local x86 toolchain, if None at init time warp will attempt to find MSVC toolchain
 
 verify_fp = False       # verify inputs and outputs are finite after each launch
 verify_cuda = False     # if true will check CUDA errors after each kernel launch / memory operation
@@ -24,3 +23,4 @@ verbose = False
 host_compiler = None    # user can specify host compiler here, otherwise will attempt to find one automatically
 
 cache_kernels = True
+kernel_cache_dir = None # path to kernel cache directory, if None a default path will be used
