@@ -26,18 +26,23 @@
 
 #if !defined(__CUDACC__)
     #define CUDA_CALLABLE
-
+    #define CUDA_CALLABLE_DEVICE
 #else
     #define CUDA_CALLABLE __host__ __device__ 
+    #define CUDA_CALLABLE_DEVICE __device__
 #endif
 
-
+#ifdef WP_VERIFY_FP
+#define FP_CHECK 1
+#else
 #define FP_CHECK 0
+#endif
 
 
 namespace wp
 {
 
+
 // numeric types (used from generated kernels)
 typedef float float32;
 typedef double float64;
@@ -54,10 +59,167 @@ typedef uint32_t uint32;
 typedef int64_t int64;
 typedef uint64_t uint64;
 
+
 // matches Python string type for constant strings
 typedef char* str;
 
 
+
+struct half;
+
+CUDA_CALLABLE half float_to_half(float x);
+CUDA_CALLABLE float half_to_float(half x);
+
+struct half
+{
+    CUDA_CALLABLE inline half() : u(0) {}
+
+    CUDA_CALLABLE inline half(float f)
+    {
+        *this = float_to_half(f);
+    }
+
+    unsigned short u;
+
+    CUDA_CALLABLE inline bool operator==(const half& h) const { return u == h.u; }
+
+    CUDA_CALLABLE inline operator float32() const { return float32(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator float64() const { return float64(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator int8() const { return int8(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator uint8() const { return uint8(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator int16() const { return int16(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator uint16() const { return uint16(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator int32() const { return int32(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator uint32() const { return uint32(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator int64() const { return int64(half_to_float(*this)); }
+    CUDA_CALLABLE inline operator uint64() const { return uint64(half_to_float(*this)); }
+};
+
+static_assert(sizeof(half) == 2, "Size of half / float16 type must be 2-bytes");
+
+typedef half float16;
+
+#if __CUDA_ARCH__
+
+CUDA_CALLABLE inline half float_to_half(float x)
+{
+#if __CUDA_ARCH__ >= 700
+    half h;
+    asm("{  cvt.rn.f16.f32 %0, %1;}\n" : "=h"(h.u) : "f"(x));  
+    return h;
+#else
+    // fp16 not supported
+    return half();
+#endif
+}
+
+CUDA_CALLABLE inline float half_to_float(half x)
+{
+#if __CUDA_ARCH__ >= 700
+    float val;
+    asm("{  cvt.f32.f16 %0, %1;}\n" : "=f"(val) : "h"(x.u));
+    return val;
+#else
+    // fp16 not supported
+    return 0.0f;
+#endif
+}
+
+#else
+
+// adapted from Fabien Giesen's post: https://gist.github.com/rygorous/2156668
+inline half float_to_half(float x)
+{
+    union fp32
+    {
+        uint32 u;
+        float f;
+
+        struct
+        {
+            unsigned int mantissa : 23;
+            unsigned int exponent : 8;
+            unsigned int sign : 1;
+        };
+    };
+
+    fp32 f;
+    f.f = x;
+
+    fp32 f32infty = { 255 << 23 };
+    fp32 f16infty = { 31 << 23 };
+    fp32 magic = { 15 << 23 };
+    uint32 sign_mask = 0x80000000u;
+    uint32 round_mask = ~0xfffu; 
+    half o;
+
+    uint32 sign = f.u & sign_mask;
+    f.u ^= sign;
+
+    // NOTE all the integer compares in this function can be safely
+    // compiled into signed compares since all operands are below
+    // 0x80000000. Important if you want fast straight SSE2 code
+    // (since there's no unsigned PCMPGTD).
+
+    if (f.u >= f32infty.u) // Inf or NaN (all exponent bits set)
+        o.u = (f.u > f32infty.u) ? 0x7e00 : 0x7c00; // NaN->qNaN and Inf->Inf
+    else // (De)normalized number or zero
+    {
+        f.u &= round_mask;
+        f.f *= magic.f;
+        f.u -= round_mask;
+        if (f.u > f16infty.u) f.u = f16infty.u; // Clamp to signed infinity if overflowed
+
+        o.u = f.u >> 13; // Take the bits!
+    }
+
+    o.u |= sign >> 16;
+    return o;
+}
+
+
+inline float half_to_float(half h)
+{
+    union fp32
+    {
+        uint32 u;
+        float f;
+
+        struct
+        {
+            unsigned int mantissa : 23;
+            unsigned int exponent : 8;
+            unsigned int sign : 1;
+        };
+    };
+
+    static const fp32 magic = { 113 << 23 };
+    static const uint32 shifted_exp = 0x7c00 << 13; // exponent mask after shift
+    fp32 o;
+
+    o.u = (h.u & 0x7fff) << 13;     // exponent/mantissa bits
+    uint32 exp = shifted_exp & o.u;   // just the exponent
+    o.u += (127 - 15) << 23;        // exponent adjust
+
+    // handle exponent special cases
+    if (exp == shifted_exp) // Inf/NaN?
+        o.u += (128 - 16) << 23;    // extra exp adjust
+    else if (exp == 0) // Zero/Denormal?
+    {
+        o.u += 1 << 23;             // extra exp adjust
+        o.f -= magic.f;             // renormalize
+    }
+
+    o.u |= (h.u & 0x8000) << 16;    // sign bit
+    return o.f;
+}
+
+
+#endif
+
+
+
+
 template <typename T>
 CUDA_CALLABLE float cast_float(T x) { return (float)(x); }
 
@@ -65,7 +227,7 @@ template <typename T>
 CUDA_CALLABLE int cast_int(T x) { return (int)(x); }
 
 template <typename T>
-CUDA_CALLABLE void adj_cast_float(T x, T& adj_x, float adj_ret) { adj_x += adj_ret; }
+CUDA_CALLABLE void adj_cast_float(T x, T& adj_x, float adj_ret) { adj_x += T(adj_ret); }
 
 template <typename T>
 CUDA_CALLABLE void adj_cast_int(T x, T& adj_x, int adj_ret) { adj_x += adj_ret; }
@@ -87,10 +249,13 @@ CUDA_CALLABLE inline void adj_int64(T, T&, int64) {}
 template <typename T>
 CUDA_CALLABLE inline void adj_uint64(T, T&, uint64) {}
 
+
 template <typename T>
-CUDA_CALLABLE inline void adj_float32(T x, T& adj_x, float32 adj_ret) { adj_x += adj_ret; }
+CUDA_CALLABLE inline void adj_float16(T x, T& adj_x, float16 adj_ret) { adj_x += T(adj_ret); }
 template <typename T>
-CUDA_CALLABLE inline void adj_float64(T x, T& adj_x, float64 adj_ret) { adj_x += adj_ret; }
+CUDA_CALLABLE inline void adj_float32(T x, T& adj_x, float32 adj_ret) { adj_x += T(adj_ret); }
+template <typename T>
+CUDA_CALLABLE inline void adj_float64(T x, T& adj_x, float64 adj_ret) { adj_x += T(adj_ret); }
 
 
 #define kEps 0.0f
@@ -121,19 +286,107 @@ inline CUDA_CALLABLE void adj_sign(int x, int adj_x, int& adj_ret) { }
 inline CUDA_CALLABLE void adj_clamp(int x, int a, int b, int& adj_x, int& adj_a, int& adj_b, int adj_ret) { }
 inline CUDA_CALLABLE void adj_floordiv(int a, int b, int& adj_a, int& adj_b, int adj_ret) { }
 
-
 // basic ops for float types
 inline CUDA_CALLABLE float mul(float a, float b) { return a*b; }
-inline CUDA_CALLABLE float div(float a, float b) { return a/b; }
+inline CUDA_CALLABLE float div(float a, float b)
+{
+#if FP_CHECK
+    if (!isfinite(a) || !isfinite(b) || b == 0.0f)
+    {
+        printf("%s:%d div(%f, %f)\n", __FILE__, __LINE__, a, b);
+        assert(0);
+    }
+#endif
+    return a/b;
+}
 inline CUDA_CALLABLE float add(float a, float b) { return a+b; }
 inline CUDA_CALLABLE float sub(float a, float b) { return a-b; }
 inline CUDA_CALLABLE float min(float a, float b) { return a<b?a:b; }
 inline CUDA_CALLABLE float max(float a, float b) { return a>b?a:b; }
-inline CUDA_CALLABLE float mod(float a, float b) { return fmodf(a, b); }
-inline CUDA_CALLABLE float log(float a) { return logf(a); }
-inline CUDA_CALLABLE float exp(float a) { return expf(a); }
-inline CUDA_CALLABLE float pow(float a, float b) { return powf(a, b); }
-inline CUDA_CALLABLE float floordiv(float a, float b) { return float(int(a/b)); }
+inline CUDA_CALLABLE float mod(float a, float b)
+{
+#if FP_CHECK
+    if (!isfinite(a) || !isfinite(b) || b == 0.0f)
+    {
+        printf("%s:%d mod(%f, %f)\n", __FILE__, __LINE__, a, b);
+        assert(0);
+    }
+#endif
+    return fmodf(a, b);
+}
+inline CUDA_CALLABLE float log(float a)
+{
+#if FP_CHECK
+    if (!isfinite(a) || a < 0.0f)
+    {
+        printf("%s:%d log(%f)\n", __FILE__, __LINE__, a);
+        assert(0);
+    }
+#endif
+    return logf(a);
+}
+
+inline CUDA_CALLABLE float log2(float a) 
+{
+#if FP_CHECK
+    if (!isfinite(a) || a < 0.0f)
+    {
+        printf("%s:%d log2(%f)\n", __FILE__, __LINE__, a);
+        assert(0);
+    }
+#endif
+
+    return log2f(a);    
+}
+inline CUDA_CALLABLE float log10(float a) 
+{
+#if FP_CHECK
+    if (!isfinite(a) || a < 0.0f)
+    {
+        printf("%s:%d log10(%f)\n", __FILE__, __LINE__, a);
+        assert(0);
+    }
+#endif
+
+    return log10f(a); 
+}
+
+
+inline CUDA_CALLABLE float exp(float a)
+{
+    float result = expf(a);
+#if FP_CHECK
+    if (!isfinite(a) || !isfinite(result))
+    {
+        printf("%s:%d exp(%f) = %f\n", __FILE__, __LINE__, a, result);
+        assert(0);
+    }
+#endif
+    return result;
+}
+inline CUDA_CALLABLE float pow(float a, float b)
+{
+    float result = powf(a, b);
+#if FP_CHECK
+    if (!isfinite(a) || !isfinite(b) || !isfinite(result))
+    {
+        printf("%s:%d pow(%f, %f) = %f\n", __FILE__, __LINE__, a, b, result);
+        assert(0);
+    }
+#endif
+    return result;
+}
+inline CUDA_CALLABLE float floordiv(float a, float b)
+{
+#if FP_CHECK
+    if (!isfinite(a) || !isfinite(b) || b == 0.0f)
+    {
+        printf("%s:%d mod(%f, %f)\n", __FILE__, __LINE__, a, b);
+        assert(0);
+    }
+#endif
+    return float(int(a/b));
+}
 
 inline CUDA_CALLABLE float leaky_min(float a, float b, float r) { return min(a, b); }
 inline CUDA_CALLABLE float leaky_max(float a, float b, float r) { return max(a, b); }
@@ -143,13 +396,29 @@ inline CUDA_CALLABLE float sign(float x) { return x < 0.0f ? -1.0f : 1.0f; }
 inline CUDA_CALLABLE float abs(float x) { return ::fabs(x); }
 inline CUDA_CALLABLE float nonzero(float x) { return x == 0.0f ? 0.0f : 1.0f; }
 
-inline CUDA_CALLABLE float acos(float x) { return ::acos(min(max(x, -1.0f), 1.0f)); }
-inline CUDA_CALLABLE float asin(float x) { return ::asin(min(max(x, -1.0f), 1.0f)); }
+inline CUDA_CALLABLE float acos(float x)
+{
+    return ::acos(min(max(x, -1.0f), 1.0f));
+}
+inline CUDA_CALLABLE float asin(float x)
+{
+    return ::asin(min(max(x, -1.0f), 1.0f));
+}
 inline CUDA_CALLABLE float atan(float x) { return ::atan(x); }
 inline CUDA_CALLABLE float atan2(float y, float x) { return ::atan2(y, x); }
 inline CUDA_CALLABLE float sin(float x) { return ::sin(x); }
 inline CUDA_CALLABLE float cos(float x) { return ::cos(x); }
-inline CUDA_CALLABLE float sqrt(float x) { return ::sqrt(x); }
+inline CUDA_CALLABLE float sqrt(float x)
+{
+#if FP_CHECK
+    if (x < 0.0f)
+    {
+        printf("%s:%d sqrt(%f)\n", __FILE__, __LINE__, x);
+        assert(0);
+    }
+#endif
+    return ::sqrt(x);
+}
 inline CUDA_CALLABLE float tan(float x) { return ::tan(x); }
 inline CUDA_CALLABLE float sinh(float x) { return ::sinhf(x);}
 inline CUDA_CALLABLE float cosh(float x) { return ::coshf(x);}
@@ -162,16 +431,75 @@ inline CUDA_CALLABLE float floor(float x) { return ::floorf(x); }
 inline CUDA_CALLABLE float ceil(float x) { return ::ceilf(x); }
 
 inline CUDA_CALLABLE void adj_mul(float a, float b, float& adj_a, float& adj_b, float adj_ret) { adj_a += b*adj_ret; adj_b += a*adj_ret; }
-inline CUDA_CALLABLE void adj_div(float a, float b, float& adj_a, float& adj_b, float adj_ret) { adj_a += adj_ret/b; adj_b -= adj_ret*(a/b)/b; }
+inline CUDA_CALLABLE void adj_div(float a, float b, float& adj_a, float& adj_b, float adj_ret)
+{
+    adj_a += adj_ret/b;
+    adj_b -= adj_ret*(a/b)/b;
+#if FP_CHECK
+    if (!isfinite(adj_a) || !isfinite(adj_b))
+    {
+        printf("%s:%d - adj_div(%f, %f, %f, %f, %f)\n", __FILE__, __LINE__, a, b, adj_a, adj_b, adj_ret);
+        assert(0);
+    }
+#endif
+}
 inline CUDA_CALLABLE void adj_add(float a, float b, float& adj_a, float& adj_b, float adj_ret) { adj_a += adj_ret; adj_b += adj_ret; }
 inline CUDA_CALLABLE void adj_sub(float a, float b, float& adj_a, float& adj_b, float adj_ret) { adj_a += adj_ret; adj_b -= adj_ret; }
 inline CUDA_CALLABLE void adj_mod(float a, float b, float& adj_a, float& adj_b, float adj_ret)
 {
     printf("adj_mod not implemented for floating point types\n");
 }
-inline CUDA_CALLABLE void adj_log(float a, float& adj_a, float adj_ret) { adj_a += (1.f/a)*adj_ret; }
+inline CUDA_CALLABLE void adj_log(float a, float& adj_a, float adj_ret)
+{
+    adj_a += (1.f/a)*adj_ret;
+#if FP_CHECK
+    if (!isfinite(adj_a))
+    {
+        printf("%s:%d - adj_log(%f, %f, %f)\n", __FILE__, __LINE__, a, adj_a, adj_ret);
+        assert(0);
+    }
+#endif
+}
+
+inline CUDA_CALLABLE void adj_log2(float a, float& adj_a, float adj_ret) 
+{ 
+    adj_a += (1.f/a)*(1.f/log(2.f))*adj_ret; 
+    
+#if FP_CHECK
+    if (!isfinite(adj_a))
+    {
+        printf("%s:%d - adj_log2(%f, %f, %f)\n", __FILE__, __LINE__, a, adj_a, adj_ret);
+        assert(0);
+    }
+#endif    
+}
+
+inline CUDA_CALLABLE void adj_log10(float a, float& adj_a, float adj_ret)
+{
+    adj_a += (1.f/a)*(1.f/log(10.f))*adj_ret; 
+    
+#if FP_CHECK
+    if (!isfinite(adj_a))
+    {
+        printf("%s:%d - adj_log10(%f, %f, %f)\n", __FILE__, __LINE__, a, adj_a, adj_ret);
+        assert(0);
+    }
+#endif
+}
+
 inline CUDA_CALLABLE void adj_exp(float a, float& adj_a, float adj_ret) { adj_a += exp(a)*adj_ret; }
-inline CUDA_CALLABLE void adj_pow(float a, float b, float& adj_a, float& adj_b, float adj_ret) { adj_a += b*pow(a, b-1.f)*adj_ret; adj_b += log(a)*pow(a, b)*adj_ret; }
+inline CUDA_CALLABLE void adj_pow(float a, float b, float& adj_a, float& adj_b, float adj_ret)
+{ 
+    adj_a += b*pow(a, b-1.f)*adj_ret;
+    adj_b += log(a)*pow(a, b)*adj_ret;
+#if FP_CHECK
+    if (!isfinite(adj_a) || !isfinite(adj_b))
+    {
+        printf("%s:%d - adj_pow(%f, %f, %f, %f, %f)\n", __FILE__, __LINE__, a, b, adj_a, adj_b, adj_ret);
+        assert(0);
+    }
+#endif
+}
 inline CUDA_CALLABLE void adj_floordiv(float a, float b, float& adj_a, float& adj_b, float adj_ret) { }
 
 inline CUDA_CALLABLE void adj_min(float a, float b, float& adj_a, float& adj_b, float adj_ret)
@@ -248,21 +576,49 @@ inline CUDA_CALLABLE void adj_abs(float x, float& adj_x, float adj_ret)
 inline CUDA_CALLABLE void adj_acos(float x, float& adj_x, float adj_ret)
 {
     float d = sqrt(1.0f-x*x);
+#if FP_CHECK
+    adj_x -= (1.0f/d)*adj_ret;
+    if (!isfinite(d) || !isfinite(adj_x))
+    {
+        printf("%s:%d - adj_acos(%f, %f, %f)\n", __FILE__, __LINE__, x, adj_x, adj_ret);        
+        assert(0);
+    }
+#else    
     if (d > 0.0f)
         adj_x -= (1.0f/d)*adj_ret;
+#endif
 }
 
 inline CUDA_CALLABLE void adj_asin(float x, float& adj_x, float adj_ret)
 {
     float d = sqrt(1.0f-x*x);
+#if FP_CHECK
+    adj_x += (1.0f/d)*adj_ret;
+    if (!isfinite(d) || !isfinite(adj_x))
+    {
+        printf("%s:%d - adj_asin(%f, %f, %f)\n", __FILE__, __LINE__, x, adj_x, adj_ret);   
+        assert(0);
+    }
+#else    
     if (d > 0.0f)
         adj_x += (1.0f/d)*adj_ret;
+#endif
 }
 
 inline CUDA_CALLABLE void adj_tan(float x, float& adj_x, float adj_ret)
 {
     float cos_x = cos(x);
+#if FP_CHECK
     adj_x += (1.0f/(cos_x*cos_x))*adj_ret;
+    if (!isfinite(adj_x) || cos_x == 0.0f)
+    {
+        printf("%s:%d - adj_tan(%f, %f, %f)\n", __FILE__, __LINE__, x, adj_x, adj_ret);
+        assert(0);
+    }
+#else    
+    if (cos_x > 0.0f)
+        adj_x += (1.0f/(cos_x*cos_x))*adj_ret;
+#endif
 }
 
 inline CUDA_CALLABLE void adj_atan(float x, float& adj_x, float adj_ret)
@@ -272,7 +628,22 @@ inline CUDA_CALLABLE void adj_atan(float x, float& adj_x, float adj_ret)
 
 inline CUDA_CALLABLE void adj_atan2(float y, float x, float& adj_y, float& adj_x, float adj_ret)
 {
-    printf("arctan2 adjoint not implemented");
+    float d = x*x + y*y;
+#if FP_CHECK
+    adj_x -= y/d*adj_ret;
+    adj_y += x/d*adj_ret;
+    if (!isfinite(adj_x) || !isfinite(adj_y) || d == 0.0f)
+    {
+        printf("%s:%d - adj_atan2(%f, %f, %f, %f, %f)\n", __FILE__, __LINE__, y, x, adj_y, adj_x, adj_ret);
+        assert(0);
+    }
+#else    
+    if (d > 0.0f)
+    {
+        adj_x -= y/d*adj_ret;
+        adj_y += x/d*adj_ret;
+    }
+#endif
 }
 
 inline CUDA_CALLABLE void adj_sin(float x, float& adj_x, float adj_ret)
@@ -304,6 +675,13 @@ inline CUDA_CALLABLE void adj_tanh(float x, float& adj_x, float adj_ret)
 inline CUDA_CALLABLE void adj_sqrt(float x, float& adj_x, float adj_ret)
 {
     adj_x += 0.5f*(1.0f/sqrt(x))*adj_ret;
+#if FP_CHECK    
+    if (!isfinite(adj_x))
+    {
+        printf("%s:%d - adj_sqrt(%f, %f, %f)\n", __FILE__, __LINE__, x, adj_x, adj_ret);
+        assert(0);
+    }
+#endif
 }
 
 inline CUDA_CALLABLE void adj_round(float x, float& adj_x, float adj_ret)
@@ -331,6 +709,9 @@ inline CUDA_CALLABLE void adj_ceil(float x, float& adj_x, float adj_ret)
     // nop
 }
 
+// basic arithmetic for fp16, this is only called for adjoint accumulation
+inline CUDA_CALLABLE half add(half a, half b) { return half(float(a)+float(b)); }
+
 
 template <typename T>
 CUDA_CALLABLE inline T select(bool cond, const T& a, const T& b) { return cond?b:a; }
@@ -443,7 +824,7 @@ inline CUDA_CALLABLE int tid()
 #endif
 }
 
-inline CUDA_CALLABLE void tid(int& i, int& j)
+inline CUDA_CALLABLE_DEVICE void tid(int& i, int& j)
 {
     const int index = tid();
 
@@ -454,7 +835,7 @@ inline CUDA_CALLABLE void tid(int& i, int& j)
     j = index%n;
 }
 
-inline CUDA_CALLABLE void tid(int& i, int& j, int& k)
+inline CUDA_CALLABLE_DEVICE void tid(int& i, int& j, int& k)
 {
     const int index = tid();
 
@@ -467,7 +848,7 @@ inline CUDA_CALLABLE void tid(int& i, int& j, int& k)
     k = index%o;
 }
 
-inline CUDA_CALLABLE void tid(int& i, int& j, int& k, int& l)
+inline CUDA_CALLABLE_DEVICE void tid(int& i, int& j, int& k, int& l)
 {
     const int index = tid();
 
@@ -494,10 +875,50 @@ inline CUDA_CALLABLE T atomic_add(T* buf, T value)
 #endif
 }
 
+template<>
+inline CUDA_CALLABLE float16 atomic_add(float16* buf, float16 value)
+{
+#if defined(WP_CPU)
+    float16 old = buf[0];
+    buf[0] += value;
+    return old;
+#elif defined(WP_CUDA)
+    //return atomicAdd(buf, value);
+    
+    /* Define __PTR for atomicAdd prototypes below, undef after done */
+    #if (defined(_MSC_VER) && defined(_WIN64)) || defined(__LP64__) || defined(__CUDACC_RTC__)
+    #define __PTR   "l"
+    #else
+    #define __PTR   "r"
+    #endif /*(defined(_MSC_VER) && defined(_WIN64)) || defined(__LP64__) || defined(__CUDACC_RTC__)*/
+   
+    half r = 0.0;
+
+    #if __CUDA_ARCH__ >= 700
+
+        asm volatile ("{ atom.add.noftz.f16 %0,[%1],%2; }\n"
+                    : "=h"(r.u)
+                    : __PTR(buf), "h"(value.u)
+                    : "memory");
+    #endif
+
+    return r;
+
+    #undef __PTR
+
+#endif // defined(WP_CUDA)
+
+}
+
+
+
+inline bool CUDA_CALLABLE isfinite(float x)
+{
+    return ::isfinite(x);
+}
 
 } // namespace wp
 
-#include "array.h"
 #include "vec2.h"
 #include "vec3.h"
 #include "vec4.h"
@@ -505,9 +926,9 @@ inline CUDA_CALLABLE T atomic_add(T* buf, T value)
 #include "mat33.h"
 #include "quat.h"
 #include "mat44.h"
-#include "matnn.h"
 #include "spatial.h"
 #include "intersect.h"
+#include "intersect_adj.h"
 #include "mesh.h"
 #include "svd.h"
 #include "hashgrid.h"
@@ -520,6 +941,41 @@ inline CUDA_CALLABLE T atomic_add(T* buf, T value)
 namespace wp
 {
 
+
+// define scalar multiplication in reverse order i.e.: s*M, individual types just implement M*s
+template <typename T>
+T mul(float s, const T& x) { return mul(x, s); }
+
+template <typename T>
+void adj_mul(float s, const T& x, float& adj_s, T& adj_x, const T& adj_ret) { adj_mul(x, s, adj_x, adj_s, adj_ret); }
+
+
+// dot for scalar types just to make some templated compile for scalar/vector
+inline CUDA_CALLABLE float dot(float a, float b) { return mul(a, b); }
+inline CUDA_CALLABLE void dot(float a, float b, float& adj_a, float& adj_b, float adj_ret) { return adj_mul(a, b, adj_a, adj_b, adj_ret); }
+
+
+template <typename T>
+CUDA_CALLABLE inline T lerp(const T& a, const T& b, float t)
+{
+    return a*(1.0-t) + b*t;
+}
+
+template <>
+CUDA_CALLABLE inline float16 lerp(const float16& a, const float16& b, float t)
+{
+    return float(a)*(1.0-t) + float(b)*t;
+}
+
+
+template <typename T>
+CUDA_CALLABLE inline void adj_lerp(const T& a, const T& b, float t, T& adj_a, T& adj_b, float& adj_t, const T& adj_ret)
+{
+    adj_a += adj_ret*(1.0-t);
+    adj_b += adj_ret*t;
+    adj_t += dot(b, adj_ret) - dot(a, adj_ret);
+}
+
 inline CUDA_CALLABLE void print(const str s)
 {
     printf("%s\n", s);
@@ -565,6 +1021,11 @@ inline CUDA_CALLABLE void print(unsigned long long i)
     printf("%llu\n", i);
 }
 
+inline CUDA_CALLABLE void print(float16 f)
+{
+    printf("%g\n", half_to_float(f));
+}
+
 inline CUDA_CALLABLE void print(float f)
 {
     printf("%g\n", f);
@@ -644,7 +1105,8 @@ inline CUDA_CALLABLE void print(spatial_matrix m)
 
 
 inline CUDA_CALLABLE void adj_print(int i, int& adj_i) { printf("%d adj: %d\n", i, adj_i); }
-inline CUDA_CALLABLE void adj_print(float i, float& adj_i) { printf("%g adj: %g\n", i, adj_i); }
+inline CUDA_CALLABLE void adj_print(float f, float& adj_f) { printf("%g adj: %g\n", f, adj_f); }
+inline CUDA_CALLABLE void adj_print(half h, half& adj_h) { printf("%g adj: %g\n", half_to_float(h), half_to_float(adj_h)); }
 inline CUDA_CALLABLE void adj_print(vec2 v, vec2& adj_v) { printf("%g %g adj: %g %g \n", v.x, v.y, adj_v.x, adj_v.y); }
 inline CUDA_CALLABLE void adj_print(vec3 v, vec3& adj_v) { printf("%g %g %g adj: %g %g %g \n", v.x, v.y, v.z, adj_v.x, adj_v.y, adj_v.z); }
 inline CUDA_CALLABLE void adj_print(vec4 v, vec4& adj_v) { printf("%g %g %g %g adj: %g %g %g %g\n", v.x, v.y, v.z, v.w, adj_v.x, adj_v.y, adj_v.z, adj_v.w); }
@@ -708,6 +1170,8 @@ inline CUDA_CALLABLE void adj_expect_near(const T& actual, const T& expected, co
     // nop
 }
 
-
 } // namespace wp
 
+// include array.h so we have the print, isfinite functions for the inner array types defined
+#include "array.h"
+#include "matnn.h"
