@@ -38,12 +38,21 @@ struct spatial_vector
     }
 };
 
+inline CUDA_CALLABLE bool operator==(const spatial_vector& a, const spatial_vector& b)
+{
+    return a.w == b.w && a.v == b.v;
+}
+
+inline bool CUDA_CALLABLE isfinite(const spatial_vector& s)
+{
+    return isfinite(s.w) && isfinite(s.v);
+}
+
 CUDA_CALLABLE inline spatial_vector operator - (spatial_vector a)
 {
     return spatial_vector(-a.w, -a.v);
 }
 
-
 CUDA_CALLABLE inline spatial_vector add(const spatial_vector& a, const spatial_vector& b)
 {
     return { a.w + b.w, a.v + b.v };
@@ -97,11 +106,25 @@ CUDA_CALLABLE inline vec3 spatial_bottom(const spatial_vector& a)
 
 inline CUDA_CALLABLE float index(const spatial_vector& v, int i)
 {
+#if FP_CHECK
+    if (i < 0 || i > 5)
+    {
+        printf("spatial_vector index %d out of bounds at %s %d\n", i, __FILE__, __LINE__);
+        assert(0);
+    }
+#endif
     return v[i];
 }
 
 inline void CUDA_CALLABLE adj_index(const spatial_vector& m, int i, spatial_vector& adj_v, int& adj_i, float adj_ret)
 {
+#if FP_CHECK
+    if (i < 0 || i > 5)
+    {
+        printf("spatial_vector index %d out of bounds at %s %d\n", i, __FILE__, __LINE__);
+        assert(0);
+    }
+#endif
     adj_v[i] += adj_ret;
 }
 
@@ -223,6 +246,18 @@ struct transform
     }    
 };
 
+inline CUDA_CALLABLE bool operator==(const transform& a, const transform& b)
+{
+    return a.p == b.p && a.q == b.q;
+}
+
+
+inline bool CUDA_CALLABLE isfinite(const transform& t)
+{
+    return isfinite(t.p) && isfinite(t.q);
+}
+
+
 CUDA_CALLABLE inline transform transform_identity()
 {
     return transform(vec3(), quat_identity());
@@ -559,9 +594,39 @@ struct spatial_matrix
 };
 
 
+inline CUDA_CALLABLE bool operator==(const spatial_matrix& a, const spatial_matrix& b)
+{
+    for (int i=0; i < 6; ++i)
+        for (int j=0; j < 6; ++j)
+            if (a.data[i][j] != b.data[i][j])
+                return false;
+
+    return true;
+}
+
+inline bool CUDA_CALLABLE isfinite(const spatial_matrix& m)
+{
+    for (int i=0; i < 6; ++i)
+        for (int j=0; j < 6; ++j)
+            if (!::isfinite(m.data[i][j]))
+                return false;
+    return true;
+}
 
 inline CUDA_CALLABLE float index(const spatial_matrix& m, int row, int col)
 {
+#if FP_CHECK
+    if (row < 0 || row > 5)
+    {
+        printf("spatial_matrix row index %d out of bounds at %s %d\n", row, __FILE__, __LINE__);
+        assert(0);
+    }
+    if (col < 0 || col > 5)
+    {
+        printf("spatial_matrix col index %d out of bounds at %s %d\n", col, __FILE__, __LINE__);
+        assert(0);
+    }
+#endif
     return m.data[row][col];
 }
 
@@ -577,6 +642,34 @@ inline CUDA_CALLABLE spatial_matrix add(const spatial_matrix& a, const spatial_m
     return out;
 }
 
+inline CUDA_CALLABLE spatial_matrix sub(const spatial_matrix& a, const spatial_matrix& b)
+{
+    spatial_matrix out;
+
+    for (int i=0; i < 6; ++i)
+        for (int j=0; j < 6; ++j)
+            out.data[i][j] = a.data[i][j] - b.data[i][j];
+
+    return out;
+}
+
+inline CUDA_CALLABLE spatial_matrix mul(const spatial_matrix& a, float b)
+{
+    spatial_matrix out;
+
+    for (int i=0; i < 6; ++i)
+    {
+        for (int j=0; j < 6; ++j)
+        {
+            for (int k=0; k < 6; ++k)
+            {
+                out.data[i][j] += a.data[i][k]*b;
+            }
+        }
+    }
+    return out;
+}
+
 
 inline CUDA_CALLABLE spatial_vector mul(const spatial_matrix& a, const spatial_vector& b)
 {
@@ -710,6 +803,13 @@ inline CUDA_CALLABLE void adj_add(const spatial_matrix& a, const spatial_matrix&
     adj_b += adj_ret;
 }
 
+inline CUDA_CALLABLE void adj_sub(const spatial_matrix& a, const spatial_matrix& b, spatial_matrix& adj_a, spatial_matrix& adj_b, const spatial_matrix& adj_ret)
+{
+    adj_a += adj_ret;
+    adj_b -= adj_ret;
+}
+
+
 inline CUDA_CALLABLE void adj_mul(const spatial_matrix& a, const spatial_vector& b, spatial_matrix& adj_a, spatial_vector& adj_b, const spatial_vector& adj_ret)
 {
     adj_a += outer(adj_ret, b);
@@ -740,6 +840,18 @@ inline CUDA_CALLABLE void adj_transform_inertia(
 
 inline void CUDA_CALLABLE adj_index(const spatial_matrix& m, int row, int col, spatial_matrix& adj_m, int& adj_row, int& adj_col, float adj_ret)
 {
+#if FP_CHECK
+    if (row < 0 || row > 5)
+    {
+        printf("spatial_matrix row index %d out of bounds at %s %d\n", row, __FILE__, __LINE__);
+        assert(0);
+    }
+    if (col < 0 || col > 5)
+    {
+        printf("spatial_matrix col index %d out of bounds at %s %d\n", col, __FILE__, __LINE__);
+        assert(0);
+    }
+#endif
     adj_m.data[row][col] += adj_ret;
 }
 
