@@ -156,13 +156,15 @@ def eval_triangles(x: wp.array(dtype=wp.vec3),
                    indices: wp.array2d(dtype=int),
                    pose: wp.array(dtype=wp.mat22),
                    activation: wp.array(dtype=float),
-                   k_mu: float,
-                   k_lambda: float,
-                   k_damp: float,
-                   k_drag: float,
-                   k_lift: float,
+                   materials: wp.array2d(dtype=float),
                    f: wp.array(dtype=wp.vec3)):
     tid = wp.tid()
+    
+    k_mu = materials[tid,0]
+    k_lambda = materials[tid,1]
+    k_damp = materials[tid,2]
+    k_drag = materials[tid,3]
+    k_lift = materials[tid,4]
 
     i = indices[tid, 0]
     j = indices[tid, 1]
@@ -393,26 +395,28 @@ def eval_triangles_contact(
     num_particles: int,                # size of particles
     x: wp.array(dtype=wp.vec3),
     v: wp.array(dtype=wp.vec3),
-    indices: wp.array(dtype=int),
+    indices: wp.array2d(dtype=int),
     pose: wp.array(dtype=wp.mat22),
     activation: wp.array(dtype=float),
-    k_mu: float,
-    k_lambda: float,
-    k_damp: float,
-    k_drag: float,
-    k_lift: float,
+    materials: wp.array2d(dtype=float),     
     f: wp.array(dtype=wp.vec3)):
 
     tid = wp.tid()
     face_no = tid // num_particles     # which face
     particle_no = tid % num_particles  # which particle
 
+    k_mu = materials[face_no,0]
+    k_lambda = materials[face_no,1]
+    k_damp = materials[face_no,2]
+    k_drag = materials[face_no,3]
+    k_lift = materials[face_no,4]
+
     # at the moment, just one particle
     pos = x[particle_no]
 
-    i = indices[face_no * 3 + 0]
-    j = indices[face_no * 3 + 1]
-    k = indices[face_no * 3 + 2]
+    i = indices[face_no, 0]
+    j = indices[face_no, 1]
+    k = indices[face_no, 2]
 
     if (i == particle_no or j == particle_no or k == particle_no):
         return
@@ -569,11 +573,12 @@ def eval_bending(
         v: wp.array(dtype=wp.vec3),
         indices: wp.array2d(dtype=int),
         rest: wp.array(dtype=float),
-        ke: float,
-        kd: float,
+        bending_properties: wp.array2d(dtype=float),
         f: wp.array(dtype=wp.vec3)):
 
     tid = wp.tid()
+    ke = bending_properties[tid,0]
+    kd = bending_properties[tid,1]
 
     i = indices[tid,0]
     j = indices[tid,1]
@@ -1066,16 +1071,6 @@ def transform_wrench(t: wp.transform, x: wp.spatial_vector):
 
     return wp.spatial_vector(w, v)
 
-@wp.func
-def transform_inverse(t: wp.transform):
-
-    p = transform_get_translation(t)
-    q = transform_get_rotation(t)
-
-    q_inv = quat_inverse(q)
-    return transform(quat_rotate(q_inv, p)*(0.0 - 1.0), q_inv)
-
-
 
 # computes adj_t^-T*I*adj_t^-1 (tensor change of coordinates), Frank & Park, section 8.2.3, pg 290
 @wp.func
@@ -1281,7 +1276,7 @@ def eval_body_joints(body_q: wp.array(dtype=wp.transform),
         axis_c = wp.transform_vector(X_wc, axis)
 
         # swing twist decomposition
-        twist = wp.quat_twist(axis, r_err)
+        twist = quat_twist(axis, r_err)
 
         q = wp.acos(twist[3])*2.0*wp.sign(wp.dot(axis, wp.vec3(twist[0], twist[1], twist[2])))
         qd = wp.dot(w_err, axis_p)
@@ -1309,7 +1304,7 @@ def eval_body_joints(body_q: wp.array(dtype=wp.transform),
         q_pc = wp.quat_inverse(q_off)*wp.quat_inverse(q_p)*q_c*q_off
         
         # decompose to a compound rotation each axis 
-        angles = wp.quat_decompose(q_pc)
+        angles = quat_decompose(q_pc)
 
         # reconstruct rotation axes
         axis_0 = wp.vec3(1.0, 0.0, 0.0)
@@ -1337,7 +1332,7 @@ def eval_body_joints(body_q: wp.array(dtype=wp.transform),
         q_pc = wp.quat_inverse(q_off)*wp.quat_inverse(q_p)*q_c*q_off
        
         # decompose to a compound rotation each axis 
-        angles = wp.quat_decompose(q_pc)
+        angles = quat_decompose(q_pc)
 
         # reconstruct rotation axes
         axis_0 = wp.vec3(1.0, 0.0, 0.0)
@@ -1458,8 +1453,8 @@ def compute_forces(model, state, particle_f, body_f):
             particle_f)
 
     # triangle elastic and lift/drag forces
-    if (model.tri_count and model.tri_ke > 0.0):
-
+    if (model.tri_count):
+            
         wp.launch(kernel=eval_triangles,
                     dim=model.tri_count,
                     inputs=[
@@ -1468,17 +1463,13 @@ def compute_forces(model, state, particle_f, body_f):
                         model.tri_indices,
                         model.tri_poses,
                         model.tri_activations,
-                        model.tri_ke,
-                        model.tri_ka,
-                        model.tri_kd,
-                        model.tri_drag,
-                        model.tri_lift
+                        model.tri_materials
                     ],
                     outputs=[particle_f],
                     device=model.device)
 
     # triangle/triangle contacts
-    if (model.enable_tri_collisions and model.tri_count and model.tri_ke > 0.0):
+    if (model.enable_tri_collisions and model.tri_count):
 
         wp.launch(kernel=eval_triangles_contact,
                     dim=model.tri_count * model.particle_count,
@@ -1489,11 +1480,7 @@ def compute_forces(model, state, particle_f, body_f):
                         model.tri_indices,
                         model.tri_poses,
                         model.tri_activations,
-                        model.tri_ke,
-                        model.tri_ka,
-                        model.tri_kd,
-                        model.tri_drag,
-                        model.tri_lift
+                        model.tri_materials
                     ],
                     outputs=[particle_f],
                     device=model.device)
@@ -1503,7 +1490,7 @@ def compute_forces(model, state, particle_f, body_f):
 
         wp.launch(kernel=eval_bending,
                     dim=model.edge_count,
-                    inputs=[state.particle_q, state.particle_qd, model.edge_indices, model.edge_rest_angle, model.edge_ke, model.edge_kd],
+                    inputs=[state.particle_q, state.particle_qd, model.edge_indices, model.edge_rest_angle, model.edge_bending_properties],
                     outputs=[particle_f],
                     device=model.device)
 
