@@ -12,8 +12,10 @@ import struct
 import zlib
 import numpy as np
 
+from typing import Any
 from typing import Tuple
-
+from typing import TypeVar
+from typing import Generic
 
 
 class constant:
@@ -23,26 +25,16 @@ class constant:
         x: Compile-time constant value, can be any of the built-in math types.    
     """
 
-    _hash = hashlib.sha256()
-    _hash_digest = None
-
     def __init__(self, x):
 
         self.val = x
-        curframe = inspect.currentframe()
-        calframe = inspect.getouterframes(curframe)
-        line_src = calframe[1][4][0]
-        constant._hash.update(bytes(line_src, 'utf-8'))
+        constant._hash.update(bytes(str(x), 'utf-8'))
 
     def __eq__(self, other):
         return self.val == other
 
-    @classmethod
-    def get_hash(cls):
-
-        if cls._hash_digest is None:
-            cls._hash_digest = cls._hash.digest()
-        return cls._hash_digest
+    # shared hash for all constants    
+    _hash = hashlib.sha256()
 
 #----------------------
 # built-in types
@@ -108,7 +100,7 @@ class transform(ctypes.Array):
     _shape_ = (7,)
     _type_ = ctypes.c_float
 
-    def __init__(self, p, q):
+    def __init__(self, p=(0.0, 0.0, 0.0), q=(0.0, 0.0, 0.0, 1.0)):
         self[0:3] = vec3(*p)
         self[3:7] = quat(*q)
 
@@ -125,6 +117,14 @@ class void:
     def __init__(self):
         pass
 
+class float16:
+
+    _length_ = 1
+    _type_ = ctypes.c_uint16
+
+    def __init__(self, x=0.0):
+        self.value = x
+
 class float32:
 
     _length_ = 1
@@ -206,7 +206,7 @@ class uint64:
         self.value = x
 
                
-scalar_types = [int8, uint8, int16, uint16, int32, uint32, int64, uint64, float32, float64]
+scalar_types = [int8, uint8, int16, uint16, int32, uint32, int64, uint64, float16, float32, float64]
 vector_types = [vec2, vec3, vec4, mat22, mat33, mat44, quat, transform, spatial_vector, spatial_matrix]
 
 
@@ -222,6 +222,7 @@ np_dtype_to_warp_type = {
     np.dtype(np.uint64): uint64,
     np.dtype(np.byte): int8,
     np.dtype(np.ubyte): uint8,
+    np.dtype(np.float16): float16,
     np.dtype(np.float32): float32,
     np.dtype(np.float64): float64
 }
@@ -281,14 +282,26 @@ class array_t(ctypes.Structure):
 
     _fields_ = [("data", ctypes.c_uint64),
                 ("shape", ctypes.c_int32*ARRAY_MAX_DIMS),
+                ("strides", ctypes.c_int32*ARRAY_MAX_DIMS),
                 ("ndim", ctypes.c_int32)]
     
     def __init__(self):
         self.data = 0
         self.shape = (0,)*ARRAY_MAX_DIMS
+        self.strides = (0,)*ARRAY_MAX_DIMS
         self.ndim = 0       
         
 
+def type_ctype(dtype):
+
+    if dtype == float:
+        return ctypes.c_float
+    elif dtype == int:
+        return ctypes.c_int32
+    else:
+        # scalar type
+        return dtype._type_
+
 def type_length(dtype):
     if (dtype == float or dtype == int):
         return 1
@@ -296,41 +309,48 @@ def type_length(dtype):
         return dtype._length_
 
 def type_size_in_bytes(dtype):
-    if (dtype == float or dtype == int):
+    if (dtype == float or dtype == int or dtype == ctypes.c_float or dtype == ctypes.c_int32):
         return 4
-    else:
-        return dtype._length_*ctypes.sizeof(dtype._type_)
+    elif hasattr(dtype, "_type_"):
+        return getattr(dtype, "_length_", 1) * ctypes.sizeof(dtype._type_)
+    else:   
+        return 0
 
-def type_ctype(dtype):
+def type_to_warp(dtype):
     if (dtype == float):
-        return ctypes.c_float
+        return float32
     elif (dtype == int):
-        return ctypes.c_int32
+        return int32
     else:
-        return dtype._type_
+        return dtype
 
-def type_typestr(ctype):
+def type_typestr(dtype):
    
-    if (ctype == ctypes.c_float):
+    if dtype == float16:
+        return "<f2"
+    elif dtype == float32:
         return "<f4"
-    elif (ctype == ctypes.c_double):
+    elif dtype == float64:
         return "<f8"
-    elif (ctype == ctypes.c_int8):
+    elif dtype == int8:
         return "b"
-    elif (ctype == ctypes.c_uint8):
+    elif dtype == uint8:
         return "B"
-    elif (ctype == ctypes.c_int16):
+    elif dtype == int16:
         return "<i2"
-    elif (ctype == ctypes.c_uint16):
+    elif dtype == uint16:
         return "<u2"
-    elif (ctype == ctypes.c_int32):
+    elif dtype == int32:
         return "<i4"
-    elif (ctype == ctypes.c_uint32):
+    elif dtype == uint32:
         return "<u4"
-    elif (ctype == ctypes.c_int64):
+    elif dtype == int64:
         return "<i8"
-    elif (ctype == ctypes.c_uint64):
+    elif dtype == uint64:
         return "<u8"
+    elif issubclass(dtype, ctypes.Array):
+        # vector types all currently float type
+        return "<f4"
     else:
         raise Exception("Unknown ctype")
 
@@ -372,10 +392,17 @@ def types_equal(a, b):
     else:
         return a == b
 
+def strides_from_shape(shape:Tuple, dtype):
+    lower_dims = np.array(shape+(1,))[1:]
+    # use 'C' (row-major) ordering by default
+    reverse_dim_prod = np.cumprod(lower_dims[::-1])[::-1]
+    return tuple(reverse_dim_prod * type_size_in_bytes(dtype))
 
-class array:
+T = TypeVar('T')
 
-    def __init__(self, data=None, dtype=None, shape=None, length=0, ptr=None, capacity=0, device=None, copy=True, owner=True, ndim=None, requires_grad=False):
+class array (Generic[T]):
+
+    def __init__(self, data=None, dtype: T=None, shape=None, strides = None, length=0, ptr=None, capacity=0, device=None, copy=True, owner=True, ndim=None, requires_grad=False):
         """ Constructs a new Warp array object from existing data.
 
         When the ``data`` argument is a valid list, tuple, or ndarray the array will be constructed from this object's data.
@@ -391,6 +418,7 @@ class array:
             data (Union[list, tuple, ndarray]) An object to construct the array from, can be a Tuple, List, or generally any type convertable to an np.array
             dtype (Union): One of the built-in types, e.g.: :class:`warp.mat33`, if dtype is None and data an ndarray then it will be inferred from the array data type
             shape (Tuple): Dimensions of the array
+            strides (Tuple): Number of bytes in each dimension between successive elements of the array
             length (int): Number of elements (rows) of the data type (deprecated, users should use `shape` argument)
             ptr (uint64): Address of an external memory address to alias (data should be None)
             capacity (int): Maximum size in bytes of the ptr allocation (data should be None)
@@ -402,6 +430,7 @@ class array:
         """
 
         self.owner = False
+        self.grad = None
 
         if shape == None:
             shape = (length,)   
@@ -439,7 +468,7 @@ class array:
 
             try:
                 # try to convert src array to destination type
-                arr = arr.astype(dtype=type_typestr(dtype._type_))
+                arr = arr.astype(dtype=type_typestr(dtype), copy=False)
             except:
                 raise RuntimeError(f"Could not convert input data with type {arr.dtype} to array with type {dtype._type_}")
             
@@ -452,6 +481,7 @@ class array:
 
             ptr = arr.__array_interface__["data"][0]
             shape = arr.__array_interface__["shape"]
+            strides = arr.__array_interface__.get("strides", None)
 
             # Convert input shape to Warp
             if type_length(dtype) > 1:
@@ -473,6 +503,10 @@ class array:
                     raise RuntimeError(f"Last dimensions of input array should match the specified data type, given shape {arr.shape}, expected last dimensions to match dtype shape {dtype._shape_}")
 
                 shape = leading_shape
+            
+                if strides is not None:
+                    strides = strides[0:-dtype_ndim]
+            
 
 
             if (device == "cpu" and copy == False):
@@ -481,6 +515,7 @@ class array:
                 self.ptr = ptr
                 self.dtype=dtype
                 self.shape=shape
+                self.strides = strides
                 self.capacity=arr.size*type_size_in_bytes(dtype)
                 self.device = device
                 self.owner = False
@@ -495,7 +530,7 @@ class array:
                 # otherwise, we must transfer to device memory
                 # create a host wrapper around the numpy array
                 # and a new destination array to copy it to
-                src = array(dtype=dtype, shape=shape, capacity=arr.size*type_size_in_bytes(dtype), ptr=ptr, device='cpu', copy=False, owner=False)
+                src = array(dtype=dtype, shape=shape, strides=strides, capacity=arr.size*type_size_in_bytes(dtype), ptr=ptr, device='cpu', copy=False, owner=False)
                 dest = empty(shape, dtype=dtype, device=device, requires_grad=requires_grad)
                 dest.owner = False
                 
@@ -512,6 +547,7 @@ class array:
             
             # explicit construction from ptr to external memory
             self.shape = shape
+            self.strides = strides
             self.capacity = capacity
             self.dtype = dtype
             self.ptr = ptr
@@ -532,24 +568,34 @@ class array:
         for d in self.shape:
             self.size *= d
 
-        # save flag, controls if gradients will be computed in by wp.Tape
-        self.requires_grad = requires_grad
+        # update byte strides and contiguous flag
+        contiguous_strides = strides_from_shape(self.shape, self.dtype)
+        if strides is None:
+            self.strides = contiguous_strides
+            self.is_contiguous = True
+        else:
+            self.strides = strides
+            self.is_contiguous = strides[:ndim] == contiguous_strides[:ndim]
 
         # store flat shape (including type shape)
         if dtype in vector_types:
             # vector type, flatten the dimensions into one tuple
             arr_shape = (*self.shape, *self.dtype._shape_)
+            dtype_strides =  strides_from_shape(self.dtype._shape_, self.dtype._type_) 
+            arr_strides = (*self.strides, *dtype_strides)
         else:
             # scalar type
             arr_shape = self.shape
+            arr_strides = self.strides
 
         # set up array interface access so we can treat this object as a numpy array
         if device == "cpu":
 
             self.__array_interface__ = { 
                 "data": (self.ptr, False), 
-                "shape": arr_shape,  
-                "typestr": type_typestr(type_ctype(self.dtype)), 
+                "shape": tuple(arr_shape),  
+                "strides": tuple(arr_strides),  
+                "typestr": type_typestr(self.dtype), 
                 "version": 3 
             }
 
@@ -558,11 +604,17 @@ class array:
 
             self.__cuda_array_interface__ = {
                 "data": (self.ptr, False),
-                "shape": arr_shape,
-                "typestr": type_typestr(type_ctype(self.dtype)),
+                "shape": tuple(arr_shape),
+                "strides": tuple(arr_strides),
+                "typestr": type_typestr(self.dtype),
                 "version": 2
             }
 
+        # controls if gradients will be computed in by wp.Tape
+        # this will trigger allocation of a gradient array if it doesn't exist already
+        self.requires_grad = requires_grad
+
+
     def __del__(self):
         
         try:
@@ -606,13 +658,29 @@ class array:
 
         for i in range(a.ndim):
             a.shape[i] = self.shape[i]
+            a.strides[i] = self.strides[i]
 
         return a        
 
+    def __setattr__(self, __name: str, __value: Any) -> None:
+
+        if __name == "requires_grad" and __value == True and self.grad == None:
+            self._alloc_grad()
+
+        return super().__setattr__(__name, __value)
+
+    def _alloc_grad(self):
+
+        from warp.context import zeros
+        self.grad = zeros(shape=self.shape, dtype=self.dtype, device=self.device, requires_grad=False)
+
 
     def zero_(self):
 
         from warp.context import runtime
+            
+        if not self.is_contiguous:
+            raise RuntimeError(f"Assigning to non-continuguous arrays is unsupported.")
 
         if (self.device == "cpu"):
             runtime.core.memset_host(ctypes.cast(self.ptr,ctypes.POINTER(ctypes.c_int)), ctypes.c_int(0), ctypes.c_size_t(self.size*type_size_in_bytes(self.dtype)))
@@ -621,6 +689,27 @@ class array:
             runtime.core.memset_device(ctypes.cast(self.ptr,ctypes.POINTER(ctypes.c_int)), ctypes.c_int(0), ctypes.c_size_t(self.size*type_size_in_bytes(self.dtype)))
 
 
+    def fill_(self, value):
+
+        from warp.context import runtime
+            
+        if not self.is_contiguous:
+            raise RuntimeError(f"Assigning to non-continuguous arrays is unsupported.")
+
+        # convert value to array type
+        src_type = type_ctype(self.dtype)
+        src_value = src_type(value)
+
+        # cast to a 4-byte integer for memset
+        dest_ptr = ctypes.cast(ctypes.pointer(src_value), ctypes.POINTER(ctypes.c_int))
+        dest_value = dest_ptr.contents
+
+        if (self.device == "cpu"):
+            runtime.core.memset_host(ctypes.cast(self.ptr,ctypes.POINTER(ctypes.c_int)), dest_value, ctypes.c_size_t(self.size*type_size_in_bytes(self.dtype)))
+
+        if(self.device == "cuda"):
+            runtime.core.memset_device(ctypes.cast(self.ptr,ctypes.POINTER(ctypes.c_int)), dest_value, ctypes.c_size_t(self.size*type_size_in_bytes(self.dtype)))
+
     # equivalent to wrapping src data in an array and copying to self
     def assign(self, src):
         from warp.context import copy
@@ -727,14 +816,14 @@ class Mesh:
         if (points.device != indices.device):
             raise RuntimeError("Mesh points and indices must live on the same device")
 
-        if (points.dtype != vec3):
-            raise RuntimeError("Mesh points should be an array of type wp.vec3")
+        if (points.dtype != vec3 or not points.is_contiguous):
+            raise RuntimeError("Mesh points should be a contiguous array of type wp.vec3")
 
-        if (velocities and velocities.dtype != vec3):
-            raise RuntimeError("Mesh velocities should be an array of type wp.vec3")
+        if (velocities and (velocities.dtype != vec3 or not velocities.is_contiguous)):
+            raise RuntimeError("Mesh velocities should be a contiguous array of type wp.vec3")
 
-        if (indices.dtype != int32):
-            raise RuntimeError("Mesh indices should be an array of type wp.int32")
+        if (indices.dtype != int32 or not indices.is_contiguous):
+            raise RuntimeError("Mesh indices should be a contiguous array of type wp.int32")
 
 
         self.device = points.device
@@ -961,3 +1050,86 @@ class HashGrid:
             pass
 
 
+class MarchingCubes:
+
+    def __init__(self, nx: int, ny: int, nz: int, max_verts: int, max_tris: int, device):
+
+        if device == "cpu":
+            raise RuntimeError("Only CUDA devices are supported for marching cubes")
+
+        self.nx = nx
+        self.ny = ny
+        self.nz = nz
+
+        self.device = device
+
+        self.max_verts = max_verts
+        self.max_tris = max_tris
+
+        from warp.context import zeros
+
+        self.verts = zeros(max_verts, dtype=vec3, device=device)
+        self.indices = zeros(max_tris*3, dtype=int, device=device)
+
+        from warp.context import runtime
+
+        # bindings to warp.so
+        self.alloc = runtime.core.marching_cubes_create_device        
+        self.alloc.restype = ctypes.c_uint64
+        self.free = runtime.core.marching_cubes_destroy_device
+
+        # alloc surfacer
+        self.id = ctypes.c_uint64(self.alloc())
+
+    def __del__(self):
+
+        # destroy surfacer
+        self.free(self.id)
+
+    def resize(self, nx: int, ny: int, nz: int, max_verts: int, max_tris: int):
+        
+        # actual allocations will be resized on next call to surface()
+        self.nx = nx
+        self.ny = ny
+        self.nz = nz
+        self.max_verts = max_verts
+        self.max_tris = max_tris
+
+
+    def surface(self, field: array(dtype=float), threshold: float):
+        
+        from warp.context import runtime
+
+        # WP_API int marching_cubes_surface_host(const float* field, int nx, int ny, int nz, float threshold, wp::vec3* verts, int* triangles, int max_verts, int max_tris, int* out_num_verts, int* out_num_tris);
+        num_verts = ctypes.c_int(0)
+        num_tris = ctypes.c_int(0)
+
+        runtime.core.marching_cubes_surface_device.restype = ctypes.c_int
+
+        error = runtime.core.marching_cubes_surface_device(self.id,
+                                                           ctypes.cast(field.ptr, ctypes.c_void_p),
+                                                           self.nx,
+                                                           self.ny,
+                                                           self.nz,
+                                                           ctypes.c_float(threshold),
+                                                           ctypes.cast(self.verts.ptr, ctypes.c_void_p),
+                                                           ctypes.cast(self.indices.ptr, ctypes.c_void_p),
+                                                           self.max_verts,
+                                                           self.max_tris,
+                                                           ctypes.c_void_p(ctypes.addressof(num_verts)),
+                                                           ctypes.c_void_p(ctypes.addressof(num_tris)))
+
+        if error:
+            raise RuntimeError("Error occured buffers may not be large enough, marching cubes required at least {num_verts} vertices, and {num_tris} triangles.")
+
+
+        # resize the geometry arrays
+        self.verts.shape = (num_verts.value, )
+        self.indices.shape = (num_tris.value*3, )
+
+        self.verts.size = num_verts.value
+        self.indices.size = num_tris.value*3
+
+
+
+
